import { Span } from 'opentracing';
import { Observable } from 'rxjs';
import { LanguageClient } from './lang-handler';
import { FileSystemEntries } from './match-files';
import { OverlayFileSystem } from './memfs';
export interface FileSystem {
    /**
     * Returns files in the workspace under base that cannot be fetched synchronously, such as remote files.
     *
     * @param base A URI under which to search, resolved relative to the rootUri
     * @return An Observable that emits URIs
     */
    getAsyncWorkspaceFiles(base?: string, childOf?: Span): Observable<string>;
    /**
     * Returns the content of a text document
     *
     * @param uri The URI of the text document, resolved relative to the rootUri
     * @return An Observable that emits the text document content
     */
    getTextDocumentContent(uri: string, childOf?: Span): Observable<string>;
    /**
     * Returns an IterableIterator for all URIs known to exist in the workspace (content loaded or not)
     */
    asyncUris(): IterableIterator<string>;
    /**
     * Returns true if the given file is known to exist in the workspace (content loaded or not)
     *
     * @param uri URI to a file
     */
    has(uri: string): boolean;
    /**
     * Returns the file content for the given URI.
     * Will throw an Error if no available in-memory.
     * Use FileSystemUpdater.ensure() to ensure that the file is available.
     */
    get(uri: string): string | undefined;
    /**
     * Adds a file to the local cache
     *
     * @param uri The URI of the file
     * @param content The optional content
     */
    add(uri: string, content?: string): void;
    getFileSystemEntries(path: string): FileSystemEntries;
}
/**
 * In-memory file cache node which represents either a folder or a file
 */
export interface FileSystemNode {
    file: boolean;
    children: Map<string, FileSystemNode>;
}
export declare abstract class InMemoryFileSystem implements FileSystem {
    readonly files: Map<string, string | undefined>;
    /**
     * File tree root
     */
    rootNode: FileSystemNode;
    /**
     * Contains a Map of all URIs that exist in the workspace, optionally with a content.
     * File contents for URIs in it do not neccessarily have to be fetched already.
     */
    constructor(files: Map<string, string | undefined>);
    /**
     * Adds a file to the local cache
     *
     * @param uri The URI of the file
     * @param content The optional content
     */
    add(uri: string, content?: string): void;
    /**
     * Called by TS service to scan virtual directory when TS service looks for source files that belong to a project
     */
    getFileSystemEntries(directory: string): FileSystemEntries;
    /**
     * Returns an IterableIterator for all URIs known to exist in the workspace (content loaded or not)
     */
    asyncUris(): IterableIterator<string>;
    /**
     * Returns true if the given file is known to exist in the workspace (content loaded or not)
     *
     * @param uri URI to a file
     */
    has(uri: string): boolean;
    /**
     * Returns the file content for the given URI.
     * Will throw an Error if no available in-memory.
     * Use FileSystemUpdater.ensure() to ensure that the file is available.
     */
    get(uri: string): string | undefined;
    /**
     * Returns all files in the workspace under base
     *
     * @param base A URI under which to search, resolved relative to the rootUri
     * @return An Observable that emits URIs
     */
    abstract getAsyncWorkspaceFiles(base?: string | undefined, childOf?: Span | undefined): Observable<string>;
    /**
     * Returns the content of a text document
     *
     * @param uri The URI of the text document, resolved relative to the rootUri
     * @return An Observable that emits the text document content
     */
    abstract getTextDocumentContent(uri: string, childOf?: Span | undefined): Observable<string>;
}
export declare class RemoteFileSystem extends InMemoryFileSystem implements FileSystem {
    private client;
    constructor(client: LanguageClient);
    /**
     * The files request is sent from the server to the client to request a list of all files in the workspace or inside the directory of the base parameter, if given.
     * A language server can use the result to index files by filtering and doing a content request for each text document of interest.
     */
    getAsyncWorkspaceFiles(base?: string, childOf?: Span): Observable<string>;
    /**
     * The content request is sent from the server to the client to request the current content of
     * any text document. This allows language servers to operate without accessing the file system
     * directly.
     */
    getTextDocumentContent(uri: string, childOf?: Span): Observable<string>;
}
export declare class LocalFileSystem implements FileSystem {
    private rootUri;
    /**
     * @param rootUri The root URI that is used if `base` is not specified
     */
    constructor(rootUri: string);
    /**
     * Converts the URI to an absolute path on the local disk
     */
    protected resolveUriToPath(uri: string): string;
    getAsyncWorkspaceFiles(base?: string): Observable<string>;
    getTextDocumentContent(uri: string): Observable<string>;
    /**
     * Returns an IterableIterator for all URIs known to exist in the workspace (content loaded or not)
     */
    asyncUris(): IterableIterator<string>;
    /**
     * Returns true if the given file is known to exist in the workspace (content loaded or not)
     *
     * @param uri URI to a file
     */
    has(uri: string): boolean;
    /**
     * Returns the file content for the given URI.
     * Will throw an Error if no available in-memory.
     * Use FileSystemUpdater.ensure() to ensure that the file is available.
     */
    get(uri: string): string | undefined;
    add(uri: string, content?: string): void;
    getFileSystemEntries(directory: string): FileSystemEntries;
}
/**
 * Synchronizes a remote file system to an in-memory file system
 *
 * TODO: Implement Disposable with Disposer
 */
export declare class FileSystemUpdater {
    private fileSystem;
    private inMemoryFs;
    /**
     * Observable for a pending or completed structure fetch
     */
    private structureFetch?;
    /**
     * Map from URI to Observable of pending or completed content fetch
     */
    private fetches;
    /**
     * Limits concurrent fetches to not fetch thousands of files in parallel
     */
    private concurrencyLimit;
    constructor(fileSystem: FileSystem, inMemoryFs: OverlayFileSystem);
    /**
     * Fetches the file content for the given URI and adds the content to the in-memory file system
     *
     * @param uri URI of the file to fetch
     * @param childOf A parent span for tracing
     * @return Observable that completes when the fetch is finished
     */
    fetch(uri: string, childOf?: Span): Observable<never>;
    /**
     * Returns a promise that is resolved when the given URI has been fetched (at least once) to the in-memory file system.
     * This function cannot be cancelled because multiple callers get the result of the same operation.
     *
     * @param uri URI of the file to ensure
     * @param childOf An OpenTracing span for tracing
     * @return Observable that completes when the file was fetched
     */
    ensure(uri: string, childOf?: Span): Observable<never>;
    /**
     * Fetches the file/directory structure for the given directory from the remote file system and saves it in the in-memory file system
     *
     * @param childOf A parent span for tracing
     */
    fetchStructure(childOf?: Span): Observable<never>;
    /**
     * Returns a promise that is resolved as soon as the file/directory structure for the given directory has been synced
     * from the remote file system to the in-memory file system (at least once)
     *
     * @param span An OpenTracing span for tracing
     */
    ensureStructure(childOf?: Span): Observable<never>;
    /**
     * Invalidates the content fetch cache of a file.
     * The next call to `ensure` will do a refetch.
     *
     * @param uri URI of the file that changed
     */
    invalidate(uri: string): void;
    /**
     * Invalidates the structure fetch cache.
     * The next call to `ensureStructure` will do a refetch.
     */
    invalidateStructure(): void;
}
