"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("mz/fs");
const opentracing_1 = require("opentracing");
const path = require("path");
const rxjs_1 = require("rxjs");
const semaphore_async_await_1 = require("semaphore-async-await");
const tracing_1 = require("./tracing");
const util_1 = require("./util");
class InMemoryFileSystem {
    /**
     * Contains a Map of all URIs that exist in the workspace, optionally with a content.
     * File contents for URIs in it do not neccessarily have to be fetched already.
     */
    constructor(files) {
        this.files = files;
        /**
         * File tree root
         */
        this.rootNode = { file: false, children: new Map() };
    }
    /**
     * Adds a file to the local cache
     *
     * @param uri The URI of the file
     * @param content The optional content
     */
    add(uri, content) {
        // Make sure not to override existing content with undefined
        if (content !== undefined || !this.files.has(uri)) {
            this.files.set(uri, content);
        }
        // Add to directory tree
        // TODO: convert this to use URIs.
        const filePath = util_1.uri2path(uri);
        const components = filePath.split(/[\/\\]/).filter(c => c);
        let node = this.rootNode;
        for (const [i, component] of components.entries()) {
            const n = node.children.get(component);
            if (!n) {
                if (i < components.length - 1) {
                    const n = { file: false, children: new Map() };
                    node.children.set(component, n);
                    node = n;
                }
                else {
                    node.children.set(component, { file: true, children: new Map() });
                }
            }
            else {
                node = n;
            }
        }
    }
    /**
     * Called by TS service to scan virtual directory when TS service looks for source files that belong to a project
     */
    getFileSystemEntries(directory) {
        const ret = { files: [], directories: [] };
        let node = this.rootNode;
        const components = directory.split(path.sep).filter(c => c);
        if (components.length !== 1 || components[0]) {
            for (const component of components) {
                const n = node.children.get(component);
                if (!n) {
                    return ret;
                }
                node = n;
            }
        }
        for (const [name, value] of node.children.entries()) {
            if (value.file) {
                ret.files.push(name);
            }
            else {
                ret.directories.push(name);
            }
        }
        return ret;
    }
    /**
     * Returns an IterableIterator for all URIs known to exist in the workspace (content loaded or not)
     */
    asyncUris() {
        return this.files.keys();
    }
    /**
     * Returns true if the given file is known to exist in the workspace (content loaded or not)
     *
     * @param uri URI to a file
     */
    has(uri) {
        return this.files.has(uri);
    }
    /**
     * Returns the file content for the given URI.
     * Will throw an Error if no available in-memory.
     * Use FileSystemUpdater.ensure() to ensure that the file is available.
     */
    get(uri) {
        return this.files.get(uri);
    }
}
exports.InMemoryFileSystem = InMemoryFileSystem;
class RemoteFileSystem extends InMemoryFileSystem {
    constructor(client) {
        super(new Map());
        this.client = client;
    }
    /**
     * The files request is sent from the server to the client to request a list of all files in the workspace or inside the directory of the base parameter, if given.
     * A language server can use the result to index files by filtering and doing a content request for each text document of interest.
     */
    getAsyncWorkspaceFiles(base, childOf = new opentracing_1.Span()) {
        return this.client
            .workspaceXfiles({ base }, childOf)
            .mergeMap(textDocuments => textDocuments)
            .map(textDocument => util_1.normalizeUri(textDocument.uri));
    }
    /**
     * The content request is sent from the server to the client to request the current content of
     * any text document. This allows language servers to operate without accessing the file system
     * directly.
     */
    getTextDocumentContent(uri, childOf = new opentracing_1.Span()) {
        return this.client
            .textDocumentXcontent({ textDocument: { uri } }, childOf)
            .map(textDocument => textDocument.text);
    }
}
exports.RemoteFileSystem = RemoteFileSystem;
class LocalFileSystem {
    /**
     * @param rootUri The root URI that is used if `base` is not specified
     */
    constructor(rootUri) {
        this.rootUri = rootUri;
    }
    /**
     * Converts the URI to an absolute path on the local disk
     */
    resolveUriToPath(uri) {
        return util_1.uri2path(uri);
    }
    getAsyncWorkspaceFiles(base = this.rootUri) {
        return rxjs_1.Observable.empty();
    }
    getTextDocumentContent(uri) {
        const filePath = this.resolveUriToPath(uri);
        return rxjs_1.Observable.fromPromise(fs.readFile(filePath, 'utf8'));
    }
    /**
     * Returns an IterableIterator for all URIs known to exist in the workspace (content loaded or not)
     */
    asyncUris() {
        return new Map().keys();
    }
    /**
     * Returns true if the given file is known to exist in the workspace (content loaded or not)
     *
     * @param uri URI to a file
     */
    has(uri) {
        return fs.existsSync(util_1.uri2path(uri));
    }
    /**
     * Returns the file content for the given URI.
     * Will throw an Error if no available in-memory.
     * Use FileSystemUpdater.ensure() to ensure that the file is available.
     */
    get(uri) {
        return fs.readFileSync(util_1.uri2path(uri), 'utf8');
    }
    add(uri, content) {
        // noop
    }
    getFileSystemEntries(directory) {
        const files = [];
        const directories = [];
        for (const name of fs.readdirSync(directory)) {
            const filePath = path.join(directory, name);
            const stat = fs.statSync(filePath);
            if (stat.isFile()) {
                files.push(name);
            }
            else if (stat.isDirectory()) {
                directories.push(name);
            }
        }
        return { files, directories };
    }
}
exports.LocalFileSystem = LocalFileSystem;
/**
 * Synchronizes a remote file system to an in-memory file system
 *
 * TODO: Implement Disposable with Disposer
 */
class FileSystemUpdater {
    constructor(fileSystem, inMemoryFs) {
        this.fileSystem = fileSystem;
        this.inMemoryFs = inMemoryFs;
        /**
         * Map from URI to Observable of pending or completed content fetch
         */
        this.fetches = new Map();
        /**
         * Limits concurrent fetches to not fetch thousands of files in parallel
         */
        this.concurrencyLimit = new semaphore_async_await_1.default(100);
    } //TODO Remove inMemoryFs as a field
    /**
     * Fetches the file content for the given URI and adds the content to the in-memory file system
     *
     * @param uri URI of the file to fetch
     * @param childOf A parent span for tracing
     * @return Observable that completes when the fetch is finished
     */
    fetch(uri, childOf = new opentracing_1.Span()) {
        // Limit concurrent fetches
        const observable = rxjs_1.Observable.fromPromise(this.concurrencyLimit.wait())
            .mergeMap(() => this.fileSystem.getTextDocumentContent(uri))
            .do(content => {
            this.concurrencyLimit.signal();
            this.inMemoryFs.add(uri, content);
        }, err => {
            this.fetches.delete(uri);
        })
            .ignoreElements()
            .publishReplay()
            .refCount();
        this.fetches.set(uri, observable);
        return observable;
    }
    /**
     * Returns a promise that is resolved when the given URI has been fetched (at least once) to the in-memory file system.
     * This function cannot be cancelled because multiple callers get the result of the same operation.
     *
     * @param uri URI of the file to ensure
     * @param childOf An OpenTracing span for tracing
     * @return Observable that completes when the file was fetched
     */
    ensure(uri, childOf = new opentracing_1.Span()) {
        return tracing_1.traceObservable('Ensure content', childOf, span => {
            span.addTags({ uri });
            return this.fetches.get(uri) || this.fetch(uri, span);
        });
    }
    /**
     * Fetches the file/directory structure for the given directory from the remote file system and saves it in the in-memory file system
     *
     * @param childOf A parent span for tracing
     */
    fetchStructure(childOf = new opentracing_1.Span()) {
        const observable = tracing_1.traceObservable('Fetch workspace structure', childOf, span => this.fileSystem
            .getAsyncWorkspaceFiles(undefined, span)
            .do(uri => {
            this.inMemoryFs.add(uri);
        }, err => {
            this.structureFetch = undefined;
        })
            .ignoreElements()
            .publishReplay()
            .refCount());
        this.structureFetch = observable;
        return observable;
    }
    /**
     * Returns a promise that is resolved as soon as the file/directory structure for the given directory has been synced
     * from the remote file system to the in-memory file system (at least once)
     *
     * @param span An OpenTracing span for tracing
     */
    ensureStructure(childOf = new opentracing_1.Span()) {
        return tracing_1.traceObservable('Ensure structure', childOf, span => this.structureFetch || this.fetchStructure(span));
    }
    /**
     * Invalidates the content fetch cache of a file.
     * The next call to `ensure` will do a refetch.
     *
     * @param uri URI of the file that changed
     */
    invalidate(uri) {
        this.fetches.delete(uri);
    }
    /**
     * Invalidates the structure fetch cache.
     * The next call to `ensureStructure` will do a refetch.
     */
    invalidateStructure() {
        this.structureFetch = undefined;
    }
}
exports.FileSystemUpdater = FileSystemUpdater;
//# sourceMappingURL=fs.js.map